# Autogenerated by Astropy's setup.py on 2018-06-01 22:12:55 UTC
from __future__ import unicode_literals
import datetime

version = "2.0.7"
githash = "d6fc1577234f308103650dfe2deb652891805941"


major = 2
minor = 0
bugfix = 7

release = True
timestamp = datetime.datetime(2018, 6, 1, 22, 12, 55)
debug = False

astropy_helpers_version = "2.0.7"

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
